@@ -25,7 +25,7 @@ ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
 # from flexmock import * is evil, keep it from doing any damage
 __all__ = ['flexmock']
-__version__ = '0.10.1'
+__version__ = '0.10.5'
 
 
 import inspect
@@ -247,10 +247,19 @@ class Expectation(object):
         if isinstance(self._mock, Mock):
             return  # no sense in enforcing this for fake objects
         allowed = self.argspec
-        # TODO(herman): fix it properly so that module mocks aren't set as methods
-        is_method = (inspect.ismethod(getattr(self._mock, self.name)) and
+        # we consider object a method for purposes or not counting "self"/"cls" as argument if:
+        #  - one of inspect.ismethod, inspect.isfunction, _isclass return True
+        #    (in Python 3 it's sometimes impossible to tell whether callable is method or not,
+        #     so we try both inspect.ismethod and inspect.isfunction; classes are callable too -
+        #     they have __init__)
+        #  - it's not a static method
+        #  - the mocked object is a module - module "methods" are in fact plain functions;
+        #    unless they're classes, which means they still have __init__
+        is_method = ((inspect.ismethod(self.original) or inspect.isfunction(self.original)
+                      or _isclass(self.original)) and
                      self.method_type is not staticmethod and
-                     not isinstance(self._mock, types.ModuleType))
+                     (not isinstance(self._mock, types.ModuleType) or
+                      _isclass(self.original)))
         args_len = len(allowed.args)
         if is_method:
             args_len -= 1
@@ -653,7 +662,14 @@ class Expectation(object):
                         type(_mock.__dict__) is dict):
                     _mock.__dict__[name] = original
                 else:
-                    setattr(_mock, name, original)
+                    if self.method_type == staticmethod and sys.version_info < (3, 0):
+                        # on some Python 2 implementations (e.g. pypy), just assigning
+                        # the original staticmethod would make it a normal method,
+                        # thus an additional "self" argument would be passed to it,
+                        # we need to explicitly cast it to staticmethod
+                        setattr(_mock, name, staticmethod(original))
+                    else:
+                        setattr(_mock, name, original)
         del self
 
 
@@ -677,7 +693,7 @@ class Mock(object):
         return self._object
 
     def __exit__(self, type, value, traceback):
-        return self
+        pass
 
     def __call__(self, *kargs, **kwargs):
         """Hack to make Expectation.mock() work with parens."""
@@ -831,7 +847,9 @@ class Mock(object):
             if method_type in SPECIAL_METHODS:
                 expectation.original_function = getattr(obj, name)
             expectation.method_type = method_type
-        override = _setattr(obj, name, types.MethodType(method_instance, obj))
+        if not _isclass(obj) or expectation.method_type in SPECIAL_METHODS:
+            method_instance = types.MethodType(method_instance, obj)
+        override = _setattr(obj, name, method_instance)
         expectation._local_override = override
         if (override and not _isclass(obj) and not isinstance(obj, Mock) and
                 hasattr(obj.__class__, name)):
@@ -930,45 +948,56 @@ class Mock(object):
                        (expected_values[0].value, return_values)))
             return return_values
 
+        def _handle_matched_expectation(expectation, runtime_self, *kargs, **kwargs):
+            if not expectation.runnable():
+                raise StateError('%s expected to be called when %s is True' %
+                                 (name, expectation._get_runnable()))
+            expectation.times_called += 1
+            expectation.verify(final=False)
+            _pass_thru = _getattr(expectation, '_pass_thru')
+            _replace_with = _getattr(expectation, '_replace_with')
+            if _pass_thru:
+                return pass_thru(expectation, runtime_self, *kargs, **kwargs)
+            elif _replace_with:
+                return _replace_with(*kargs, **kwargs)
+            return_values = _getattr(expectation, 'return_values')
+            if return_values:
+                return_value = return_values[0]
+                del return_values[0]
+                return_values.append(return_value)
+            else:
+                return_value = ReturnValue()
+            if return_value.raises:
+                if _isclass(return_value.raises):
+                    raise return_value.raises(
+                        *return_value.value['kargs'], **return_value.value['kwargs'])
+                else:
+                    raise return_value.raises
+            else:
+                return return_value.value
+
         def mock_method(runtime_self, *kargs, **kwargs):
             arguments = {'kargs': kargs, 'kwargs': kwargs}
             expectation = FlexmockContainer.get_flexmock_expectation(
                 self, name, arguments)
             if expectation:
-                if not expectation.runnable():
-                    raise StateError('%s expected to be called when %s is True' %
-                                     (name, expectation._get_runnable()))
-                expectation.times_called += 1
-                expectation.verify(final=False)
-                _pass_thru = _getattr(expectation, '_pass_thru')
-                _replace_with = _getattr(expectation, '_replace_with')
-                if _pass_thru:
-                    return pass_thru(expectation, runtime_self, *kargs, **kwargs)
-                elif _replace_with:
-                    return _replace_with(*kargs, **kwargs)
-                return_values = _getattr(expectation, 'return_values')
-                if return_values:
-                    return_value = return_values[0]
-                    del return_values[0]
-                    return_values.append(return_value)
-                else:
-                    return_value = ReturnValue()
-                if return_value.raises:
-                    if _isclass(return_value.raises):
-                        raise return_value.raises(
-                            *return_value.value['kargs'], **return_value.value['kwargs'])
-                    else:
-                        raise return_value.raises
-                else:
-                    return return_value.value
-            else:
-                # make sure to clean up expectations to ensure none of them
-                # interfere with the runner's error reporing mechanism
-                # e.g. open()
-                for _, expectations in FlexmockContainer.flexmock_objects.items():
-                    for expectation in expectations:
-                        _getattr(expectation, 'reset')()
-                raise MethodSignatureError(_format_args(name, arguments))
+                return _handle_matched_expectation(expectation, runtime_self, *kargs, **kwargs)
+            # inform the user which expectation(s) for the method were _not_ matched
+            expectations = [
+                e for e in reversed(FlexmockContainer.flexmock_objects.get(self, []))
+                if e.name == name
+            ]
+            error_msg = _format_args(name, arguments)
+            if expectations:
+                for e in expectations:
+                    error_msg += '\nDid not match expectation %s' % _format_args(name, e.args)
+            # make sure to clean up expectations to ensure none of them
+            # interfere with the runner's error reporing mechanism
+            # e.g. open()
+            for _, expectations in FlexmockContainer.flexmock_objects.items():
+                for expectation in expectations:
+                    _getattr(expectation, 'reset')()
+            raise MethodSignatureError(error_msg)
 
         return mock_method
 
@@ -1215,9 +1244,13 @@ def _hook_into_pytest():
             ret = saved(item, when, **kwargs)
             if when != 'call' and ret.excinfo is None:
                 return ret
-            teardown = runner.CallInfo(flexmock_teardown, when=when)
-            teardown.result = None
-            if ret.excinfo is not None and teardown.excinfo is None:
+            if hasattr(runner.CallInfo, "from_call"):
+                teardown = runner.CallInfo.from_call(flexmock_teardown, when=when)
+                teardown.duration = ret.duration
+            else:
+                teardown = runner.CallInfo(flexmock_teardown, when=when)
+                teardown.result = None
+            if ret.excinfo is not None:
                 teardown.excinfo = ret.excinfo
             return teardown
         runner.call_runtest_hook = call_runtest_hook
